/*
 * Created on Apr 17, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package Utils;

import java.util.*;

import DataProcess.*;


/**
 * @author Lior Berry
 * UBC
*/
public class ClusterAlgo {

	protected TimeSeries 	mTimeSeries;
	protected DataSet 		mDataSet;
	protected ArrayList 	mBins;
	protected Vector		mClusterPool;
	protected Cluster		mDendogramRoot = null;
	protected int 			mNumClusters = 5;
	
	
	public static int		ALGO_RMSD		= 0;
	public static int		ALGO_NORM_RMSD	= 1;
	public static int		ALGO_NORM2_RMSD	= 2;
	
	protected int			mAlgoType = ALGO_RMSD;
	
	protected ClusterProgresCB mProgressCallBack= null;
	
	protected Vector		mCachedClusterList = null;
	
	public double GetSeriesMaxValue()
	{
		double max = 0.0;
		if (mAlgoType == ALGO_RMSD){
			max = mTimeSeries.GetMax();
		}
		else if (mAlgoType == ALGO_NORM_RMSD || mAlgoType == ALGO_NORM2_RMSD)
		{
			max = 1.0;
		}
		
		return max;
	}
	
	public double GetSeriesMinValue()
	{
		double min = 0.0;
		if (mAlgoType == ALGO_RMSD){
			min = mTimeSeries.GetMin();
		}
		else if (mAlgoType == ALGO_NORM_RMSD || mAlgoType == ALGO_NORM2_RMSD)
		{
						min = 0.0;  
		}
		
		return min;	
	}
	
	public void SetNumClusters(int num)
	{
		mNumClusters = num;
	}
	
	public int GetNumClusters()
	{
		return mNumClusters;	
	}

	class DistancePair
	{

		public double 	mDistance;
		public Cluster	mCluster;

		public DistancePair(double dist,Cluster clust)
		{
			mDistance = dist;
			mCluster = clust;
		}
	}

	public class Cluster{
		
		public double[] 	mValues 		= null;
		public Vector 		mChildClusters 	= null;
		public int[] 		mBInIndex; 		// for base clusters
		public int			mClusteringStage = -1;
		
		
		public Vector		mDistanceVec	= null;
		DistancePair		mMinDistPair	= null;
		
		public Cluster()
		{
			mDistanceVec = new Vector();
		}
		
		public void DumpBins()
		{
			String binIndices = "";
			for (int i=0;i<mBInIndex.length;i++){
				binIndices += mBInIndex[i]+",";
			}
			
			System.out.println("("+binIndices+")");
		}
		
		public void MarkBins(ArrayList bins,int id)
		{
			int binIndex;
			BinData binData;
			for (int i=0;i<mBInIndex.length;i++){
				binIndex = mBInIndex[i];
				binData = (BinData)bins.get(binIndex);
				binData.mClusterID = id;
			}
			
		}
		
		public void Merge(Cluster a, Cluster b)
		{
			mChildClusters = new Vector();
			mChildClusters.add(a);
			mChildClusters.add(b);

			// create the new Cluster values
			int len 	= a.mValues.length;
			int maxLen 	= a.mValues.length;
			Cluster maxLenCluster = null;
			
			if (a.mValues.length != b.mValues.length){
				len = Math.min(a.mValues.length,b.mValues.length);
				maxLen = Math.max(a.mValues.length,b.mValues.length);
				if (a.mValues.length > b.mValues.length){
					maxLenCluster = a;
				}
				else{
					maxLenCluster = b;
				}
			}

			// Average the vlaues across the two clusters
			mValues = new double[maxLen];
			for (int i=0;i<len;i++){
				mValues[i] = (a.mValues[i] + b.mValues[i]) / 2;			
			}

			// Handle the case where the two clusters are of different lengths		
			if (maxLenCluster != null){
				for (int i=len;i<maxLenCluster.mValues.length;i++){
					mValues[i] = maxLenCluster.mValues[i];			
				}
			}
			
			
			// Merge the stored bins
			int numBins = a.mBInIndex.length + b.mBInIndex.length;
			mBInIndex = new int[numBins];
			
			int k=0;
			for (int i=0;i<a.mBInIndex.length;i++)
				mBInIndex[k++] = a.mBInIndex[i];
				
			for (int i=0;i<b.mBInIndex.length;i++)
				mBInIndex[k++] = b.mBInIndex[i];
		}
		
		public double GetMax()
		{
			double max = 0.0;
			for(int i=0;i<mValues.length;i++){
				 if (mValues[i] > max || i == 0)
				 	max = mValues[i];
			}
			
			return max;
		}
	
		public double GetMin()
		{
			double min = 0.0;
			for(int i=0;i<mValues.length;i++){
				 if (mValues[i] < min || i == 0)
					min = mValues[i];
			}
			
			return min;
		}
		
		public void Normalize()
		{
			double max = GetMax();
			for(int i=0;i<mValues.length;i++){
				mValues[i] = mValues[i] / max;
			}
		}
		
		public void NormalizeMinMax()
		{
			double max = GetMax();
			double min = GetMin();
			
			for(int i=0;i<mValues.length;i++){
				mValues[i] = (mValues[i] - min) / (max-min);
			}
		}		
				
		public double ComputeDistance(Cluster other)
		{
			double ret = 0.0;
			double aux = 0.0;
			int len = 0;
			
			len = mValues.length;
			if (mValues.length != other.mValues.length){
				len = Math.min(mValues.length,other.mValues.length);
			}
			
			for (int i=0;i<len;i++)
			{
				aux = (mValues[i]-other.mValues[i]);
				ret += aux * aux;
				ret = Math.sqrt(aux) / len;								
			}
			
			return ret;
		}
		
		public void RegisterDistance(double dist,Cluster other)
		{
			DistancePair newDistPair = new DistancePair(dist,other);
			mDistanceVec.add(newDistPair);
			if (mMinDistPair != null && mMinDistPair.mDistance > dist){
				mMinDistPair = newDistPair;
			}
		}
		
		public void UnRegisterDistance(Cluster other)
		{
			DistancePair auxPair;
			for (int i=0;i<mDistanceVec.size();i++){
				auxPair =  (DistancePair)(mDistanceVec.elementAt(i));
				if (auxPair.mCluster == other){
					mDistanceVec.remove(auxPair);
					break;		
				}
			}
			
			mMinDistPair = null;
		}
		
		public DistancePair GetMinDistancPair()
		{
			if (mMinDistPair == null){
				DistancePair auxPair;
				for (int i=0;i<mDistanceVec.size();i++){
					auxPair =  (DistancePair)(mDistanceVec.elementAt(i));
					if (mMinDistPair == null || mMinDistPair.mDistance > auxPair.mDistance){
						mMinDistPair = auxPair;				
					}
				}
			}
			
			return 	mMinDistPair;
		}
		
	}

	public void SetData(DataSet inDataSet,
						ArrayList inBins,
						TimeSeries inTimeSeries)
	{
		mDataSet 	= inDataSet;
		mBins		= inBins;
		mTimeSeries = inTimeSeries;
	}
	
	public void SetCallBack(ClusterProgresCB inCB)
	{
		mProgressCallBack = inCB;
	}
	
	public void Reset()
	{
		mDataSet 	= null;
		mBins		= null;
		mTimeSeries = null;
		mClusterPool = null;
		mDendogramRoot = null;
		mCachedClusterList = null;
		mAlgoType = ALGO_RMSD;
	}
	
	public void SetAlgoType(int type)
	{
		mAlgoType = type;
	}
	
	public int GetAlgoType()
	{
		return mAlgoType;
	}
	
	public void Process()
	{
		PrepareInitialClusters();
		boolean finish = false;
		int stage = 0;
		int totalNumStages = mClusterPool.size()-1;
		while(!finish){
			System.out.println("Cluster pool size = "+mClusterPool.size());

			// Find the closest clusters
			DistancePair minDistPair = null;
			Cluster		 clusterA = null;
			for (int i=0;i<mClusterPool.size();i++){
				Cluster auxCluster = (Cluster)mClusterPool.elementAt(i);
				DistancePair auxDist = auxCluster.GetMinDistancPair();
				if (minDistPair == null || auxDist.mDistance < minDistPair.mDistance){
					clusterA = auxCluster;
					minDistPair = auxDist;
				}
			}
			
			// make a new cluster out of them and remove them
			Cluster newCluster = new Cluster();
			newCluster.Merge(clusterA,minDistPair.mCluster);
			
			if (mAlgoType == ALGO_NORM_RMSD){
				newCluster.Normalize();
			}
			else if (mAlgoType == ALGO_NORM2_RMSD){
				newCluster.NormalizeMinMax();
			}

			newCluster.mClusteringStage = stage;
			stage++;

			// remove old clusters
			mClusterPool.remove(clusterA);
			mClusterPool.remove(minDistPair.mCluster);

			// Compute distances from the new cluster
			
			for (int i=0;i<mClusterPool.size();i++){
				Cluster auxCluster = (Cluster)mClusterPool.elementAt(i);
				double dist = auxCluster.ComputeDistance(newCluster);
				
				auxCluster.UnRegisterDistance(clusterA);
				auxCluster.UnRegisterDistance(minDistPair.mCluster);
				
				auxCluster.RegisterDistance(dist,newCluster);
				newCluster.RegisterDistance(dist,auxCluster);
			}			

			//  add new cluster
			mClusterPool.add(newCluster);
			
			if (mClusterPool.size() == 1)
				finish = true;


			mProgressCallBack.ProgressUpdate(stage,totalNumStages);

		}
		
		mDendogramRoot = (Cluster)mClusterPool.elementAt(0);
	}
	
	public void ClassifyBins(ArrayList bins)
	{
		Vector clusterList = ExtractClusters();

		// print out the list of clusters and mark bins
		Cluster auxCluster = null;

		for (int k=0;k<clusterList.size();k++){
			auxCluster = (Cluster)clusterList.elementAt(k);
			auxCluster.DumpBins();
			auxCluster.MarkBins(bins,k);
		}
	}
	
	public Vector ExtractClusters()
	{
		if (mDendogramRoot == null)
			return null;
			
		if (mCachedClusterList != null && mCachedClusterList.size() == mNumClusters)
			return mCachedClusterList;
		
		int numBranching = mNumClusters-1;
		Vector clusterList = new Vector();
		clusterList.add(mDendogramRoot);
		
		for (int i=0;i<numBranching;i++){
			// look for the cluster with the latest clustering stage
			
			int maxClusterStage = 0;
			Cluster branchCluster = null;
			Cluster auxCluster = null;
			
			for (int k=0;k<clusterList.size();k++){
				auxCluster = (Cluster)clusterList.elementAt(k);
				if (k == 0 || maxClusterStage < auxCluster.mClusteringStage){
					branchCluster = auxCluster;
					maxClusterStage = auxCluster.mClusteringStage;
				}
			}
			
			clusterList.add(branchCluster.mChildClusters.elementAt(0));
			clusterList.add(branchCluster.mChildClusters.elementAt(1));
			clusterList.remove(branchCluster);
		}

		return clusterList;
		
	}

	private void PrepareInitialClusters()
	{
		mClusterPool = new Vector();
		// Start by creating a cluster for each bin
		
		for (int i=0;i<mBins.size();i++){
			
			Cluster nCluster = new Cluster();
			BinData data = (BinData)mBins.get(i);

			int binSize = data.mEndIntervalIndex - data.mStartIntervalIndex + 1;
			nCluster.mValues = new double[binSize];
			int j=0;
			for (int k=data.mStartIntervalIndex;k<=data.mEndIntervalIndex;k++,j++){
				nCluster.mValues[j] = mTimeSeries.GetValueByIndex(k);			 
			}
			
			nCluster.mBInIndex = new int[1];
			nCluster.mBInIndex[0] = i;
			
			if (mAlgoType == ALGO_NORM_RMSD){
				nCluster.Normalize();
			}
			else if (mAlgoType == ALGO_NORM2_RMSD){
				nCluster.NormalizeMinMax();
			}
			
			mClusterPool.add(nCluster);
		}
		
		// Compute all distances between all clusters
		
		for (int i=0;i<mClusterPool.size();i++){
			
			Cluster a = (Cluster)mClusterPool.elementAt(i);
			
			for (int j=i+1;j<mClusterPool.size();j++){
				
				Cluster b = (Cluster)mClusterPool.elementAt(j);
				double dist = a.ComputeDistance(b);
				a.RegisterDistance(dist,b);
				b.RegisterDistance(dist,a);
			}
		}

	}
	
}
